"""Basic plotting procedures."""


def plot_spectrum(mz, intensity, show=True, **kwds):
    """A simple stem-plot stick spectrum.
    
    You need to have matplotlib installed for this method to work.
    
    Arguments:
        MZ (iterable): mass over charge values.
        I (iterable): intensities corresponding to mass over charge ratios.
        show (bool): show the plot
    """
    import matplotlib.pyplot as plt
    plt.stem(mz, intensity, markerfmt=' ', use_line_collection=True, **kwds)
    if show:
        plt.show()


def plot3d(x, y, z, show=True, **kwds):
    """Make a 3D plot of data."""
    from mpl_toolkits.mplot3d import Axes3D
    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')
    Axes3D.scatter(ax, xs=x, ys=y, zs=z)
    ax.set_xlabel('x')
    ax.set_ylabel('y')
    ax.set_zlabel('z')
    if show:
        plt.show()


def plot2d(X, x_axis, x_i, x_l, y_axis, y_i, y_l, show=True, **kwds):
    import matplotlib.pyplot as plt
    plt.imshow(X, origin='lower', cmap='inferno', aspect='auto', **kwds)
    plt.xticks(x_i, x_l)
    plt.yticks(y_i, y_l)
    plt.xlabel(x_axis)
    plt.ylabel(y_axis)
    plt.tight_layout()
    if show:
        plt.show()


"""

Log:
TimsVaex cannot be a subclass vaex.dataset.HDF5MappableStorage because of potential clashes between column names and methods.
"""

import functools
import numpy as np
import pathlib

from .sql import tables_names, table2df
from .plot import plot2d


round10 = lambda x: 10**np.floor(np.log10(x)).astype(int)


class TimsVaex(object):
    def __init__(self, path2hdf, path2tdf):
        """
        Args:
            path2hdf (str): Path to the hdf5 stored timsTOF dataset.
            path2tdf (str): Path to the 'analysis.tdf' sqlite3 DB.
        """
        import vaex

        self.path2hdf = pathlib.Path(path2hdf)
        self.path2tdf = pathlib.Path(path2tdf)
        self.df = vaex.open(str(self.path2hdf))
        self.columns = tuple(self.df.columns)
        self.frames = self.table2df('frames')
        self.min_frame = self.frames.Id.min()
        self.max_frame = self.frames.Id.max()
        self.frames_no = self.max_frame-self.min_frame+1
        self._ms1_mask = self.frames.MsMsType.values == 0
        self.ms1_frames = self.frames.Id[self._ms1_mask].values
        self.retention_time = self.frames.Time


    def tables_names(self):
        """List names of tables in the SQLite db.

        Returns:
            pd.DataTable: table with names of tables one can get with 'table2df'.
        """
        return tables_names(self.path2tdf)


    def table2df(self, name):
        """Retrieve a table with SQLite connection from a data base.

        Args:
            name (str): Name of the table to extract.
        Returns:
            pd.DataFrame: required data frame.
        """
        return table2df(self.path2tdf, name)

    def __repr__(self):
        return f"{self.__class__.__name__}.df\n{repr(self.df)}"

    @functools.lru_cache(maxsize=10)
    def minmax(self, column):
        if column == 'frame':
            return self.min_frame, self.max_frame 
        else:
            return self.df[column].minmax()
 
    def min(self, column):
        return self.minmax(column)[0]

    def max(self, column):
        return self.minmax(column)[1]

    @functools.lru_cache(2)
    def intensity_per_frame(self, recalibrated=True):
        if recalibrated:
            return self.frames.SummedIntensities.values 
        else:
            import vaex
            frame2summedIntensity = self.df.groupby(by='frame',
             agg={'SummedIntensity': vaex.agg.sum('intensity')}).to_pandas_df().sort_values('frame')
            return frame2summedIntensity.SummedIntensity.values

    def plot_TIC(self, recalibrated=True, show=True):
        """Plot peak counts per frame.

        Arguments:
            recalibrated (bool): Use Bruker recalibrated total intensities or calculate them from scratch with OpenTIMS?
            show (bool): Show the plot immediately, or just add it to the canvas?
        """
        import matplotlib.pyplot as plt
        MS1 = self._ms1_mask
        I = self.intensity_per_frame(recalibrated)
        plt.plot(self.retention_time[ MS1], I[ MS1], label="MS1")
        plt.plot(self.retention_time[~MS1], I[~MS1], label="MS2")
        plt.legend()
        plt.xlabel("Retention Time")
        plt.ylabel("Intensity")
        plt.title("Total Intensity [Ion Current]")
        if show:
            plt.show()

    def intensity_given_mz_inv_ion_mobility(self,
                                            frames=None,
                                            mz_bin_borders=np.linspace(500, 2500, 1001),
                                            inv_ion_mobility_bin_borders=np.linspace(0.8, 1.7, 101)):
        if frames is None:
            frames = self.ms1_frames




    # @property
    # @functools.lru_cache(1)
    # def TIC_frame_scan(self):
    #     f,F = self.frame_minmax
    #     s,S = self.scan_minmax
    #     return self.df.sum(self.df.intensity,
                           # binby=[self.df.scan, self.df.frame],
                           # limits=[[s-.5,S+.5], [f-.5, F+.5]],
                           # shape=[S-s+1, F-f+1])

    # def TIC_plot(self, show=True):
    #     import matplotlib.pyplot as plt
    #     tic = self.TIC()
    #     plt.plot(self.ms1_frames, tic[self.F.MsMsType == 0], label='MS1')
    #     plt.plot(self.ms2_frames, tic[self.F.MsMsType != 0], label='MS2')
    #     plt.legend()
    #     plt.tight_layout()
    #     if show:
    #         plt.show()

    # def TIC_frame_scan_plot_MS1(self, show=True):
    #     s, S = self.scan_minmax
    #     I = self.TIC_frame_scan[:, self.ms1_frames-1]
    #     x_i = np.arange(0, I.shape[1], round10(I.shape[1]), dtype=int)
    #     x_l = self.ms1_frames[x_i]
    #     y_i = np.arange(0, I.shape[0], round10(I.shape[0]), dtype=int)
    #     y_l = np.arange(s, S, round10(I.shape[0]), dtype=int)
    #     plot2d(I, 'Frame', x_i, x_l, 'Scan', y_i, y_l, show)

    # def TIC_frame_scan_plot_MS2(self, show=True):
    #     s, S = self.scan_minmax
    #     I = self.TIC_frame_scan[:, self.ms2_frames-1]
    #     x_i = np.arange(0, I.shape[1], round10(I.shape[1]), dtype=int)
    #     x_l = self.ms2_frames[x_i]
    #     y_i = np.arange(0, I.shape[0], round10(I.shape[0]), dtype=int)
    #     y_l = np.arange(s, S, round10(I.shape[0]), dtype=int)
    #     plot2d(I, 'Frame', x_i, x_l, 'Scan', y_i, y_l, show)



import argparse

from timspy.df import TimsPyDF, all_columns


P = argparse.ArgumentParser(description='Convert timsTOF Pro TDF into HDF5 with another set of parameters.',
                            formatter_class=argparse.ArgumentDefaultsHelpFormatter)

AA = P.add_argument

AA('source', help='Path to the input .d folder containing TIMS data ("analysis.tdf" and "analysis.tdf_raw").')
AA('target', help='Path to the output HDF5 file.')
AA('--compression',
    help='Compression algorithm.',
    choices=('gzip','lzf','szip', 'none'),
    default='gzip')
AA('--compression_level',
    help='Compression level, valid only for gzip compression.',
    default=4,
    type=int)
AA('--chunksNo',
    help='Number of chunks. 0 for auto.',
    default=0,
    type=int)
AA('--shuffle',
    help='Perform byte shuffling to help compression.',
    action='store_true')
AA('--columns',
    help="A list of columns to be included in the output HDF5 file.",
    nargs="+",
    choices=all_columns,
    default="frame scan tof intensity")
AA('--silent', help="Do not display progress bar.", action="store_true")
args = P.parse_args()

if isinstance(args.columns, str):
    args.columns = args.columns.split()

DF = TimsPyDF(args.source)
DF.to_hdf(target_path=self.target,
          columns=args.columns,
          compression=args.compression,
          compression_level=args.compression_level,
          shuffle=args.shuffle,
          chunks=True if args.chunksNo == 0 else args.chunksNo,
          silent=args.silent)


from math import ceil
import numpy as np

from .df import TimsPyDF


class TimsPyDIA(TimsPyDF):
    def __init__(self, analysis_directory):
        """Create an instance of the TimsPyDF.

        The frames frame contains additional information about the windows.

        Args:
            analysis_directory (str, unicode string): path to the folder containing 'analysis.tdf' and 'analysis.tdf_raw'.
        """
        super().__init__(analysis_directory)
        W = self.table2df('DiaFrameMsMsWindows')
        W['mz_left'] = W.IsolationMz - W.IsolationWidth/2.0
        W['mz_right'] = W.IsolationMz + W.IsolationWidth/2.0
        self.windows = W
        
        div = ceil( len(self.frames) / (len(self.windows) + 1) )
        ms2frame2window = np.tile(np.insert(self.windows.index.values,0,-100), div) # -100 stands for MS1
        self.frames['Windows_idx'] = ms2frame2window[0:len(self.frames)]
        self.frames.Windows_idx.replace(-100, np.nan, inplace=True)


    def frames_meta(self):
        """Return meta information on every frame.

        Returns:
            pd.DataFrame: Information on every frame, including which window it belongs to.
        """
        return self.frames.merge(self.windows,
                                 left_on='Windows_idx',
                                 right_index=True,
                                 how='left')


import pandas as pd
import sqlite3


def _sql2df(path, query):
    with sqlite3.connect(path) as conn:
        return pd.read_sql_query(query, conn)


def tables_names(path):
    """List names of tables in the SQLite db.

    Arguments:
        path (str): Path to the sqlite db.

    Returns:
        pd.DataTable: table with names of tables one can get with 'table2df'.
    """
    sql = "SELECT name FROM sqlite_master WHERE TYPE = 'table'"
    return _sql2df(path, sql)


def table2df(path, name):
    """Retrieve a table with SQLite connection from a data base.

    This function is simply great for injection attacks (:
    Don't do that.

    Args:
        name (str): Name of the table to extract.
    Returns:
        pd.DataFrame: required data frame.
    """
    return _sql2df(path, f"SELECT * FROM {name}")



# This Python file uses the following encoding: utf-8
from setuptools import setup, find_packages

setup(  name='timspy',
        packages=find_packages(),
        version='0.9.1',
        description='TimsPy: access for raw timsTOF Pro data for data scientists',
        long_description='TimsPy facilitates access to the raw data gathered by timsTOF Pro mass spectrometer. Directly see data in the long format in the familiar Pandas DataFrame object.',
        author='MatteoLacki',
        author_email='matteo.lacki@gmail.com',
        url='https://github.com/MatteoLacki/timspy.git',
        keywords=['timsTOFpro', 'numpy', 'data science','mass spectrometry'],
        classifiers=['Development Status :: 1 - Planning',
                     'License :: OSI Approved :: BSD License',
                     'Intended Audience :: Science/Research',
                     'Topic :: Scientific/Engineering :: Chemistry',
                     'Programming Language :: Python :: 3.6',
                     'Programming Language :: Python :: 3.7',
                     'Programming Language :: Python :: 3.8'],
        install_requires=['pandas',
                          'opentimspy',
                          'matplotlib',
                          'opentims_bruker_bridge',
                          'tqdm'],
        extras_require={
            'vaex': ['vaex-core',
                     'vaex-hdf5',
                     'h5py'],
        },
        scripts = [
            'bin/tims2hdf5.py'
        ]
)



import numpy as np
import pandas as pd
import pathlib
import tqdm

from opentimspy.opentims import OpenTIMS, all_columns, all_columns_dtype

from .sql import tables_names, table2df


column2dtype = dict(zip(all_columns, all_columns_dtype))


class TimsPyDF(OpenTIMS):
    """TimsData that uses info about Frames."""
    def __init__(self, analysis_directory):
        """Create an instance of the TimsPyDF.

        Args:
            analysis_directory (str, unicode string): path to the folder containing 'analysis.tdf' and 'analysis.tdf_raw'.
        """
        super().__init__(analysis_directory)
        self.frames = self.table2df("frames").sort_values('Id')
        self.frames_no = self.max_frame-self.min_frame+1
        self._ms1_mask = np.zeros(self.frames_no,
                                  dtype=bool)
        self._ms1_mask[self.ms1_frames-1] = True
        self.retention_time = self.frames.Time.values


    def tables_names(self):
        """List names of tables in the SQLite db.

        Returns:
            pd.DataTable: table with names of tables one can get with 'table2df'.
        """
        return tables_names(self.analysis_directory/'analysis.tdf')


    def table2df(self, name):
        """Retrieve a table with SQLite connection from a data base.

        Args:
            name (str): Name of the table to extract.
        Returns:
            pd.DataFrame: required data frame.
        """
        return table2df(self.analysis_directory/'analysis.tdf', name)

 
    def __repr__(self):
        return f"{self.__class__.__name__}({self.peaks_cnt} peaks)"


    def summary(self):
        """Print a short summary of the data content.

        Includes the number of peaks, the minimal and the maximal frame numbers.
        """
        print(f"Peaks Inside:   {self.peaks_cnt}")
        print(f"Minimal Frame:  {self.min_frame}")
        print(f"Maximal Frame:  {self.max_frame}")


    def query(self, frames, columns=all_columns):
        """Get data from a selection of frames.

        Args:
            frames (int, iterable): Frames to choose. Passing an integer results in extracting that one frame.
            columns (tuple): which columns to extract? Defaults to all possible columns.
        Returns:
            pd.DataFrame: Data frame filled with columns with raw data.
        """
        return pd.DataFrame(super().query(frames, columns))


    def plot_peak_counts(self, show=True):
        """Plot peak counts per frame.

        Arguments:
            show (bool): Show the plot immediately, or just add it to the canvas.
        """
        import matplotlib.pyplot as plt
        MS1 = self._ms1_mask
        NP = self.frames.NumPeaks
        plt.plot(self.retention_time[ MS1], NP[ MS1], label="MS1")
        plt.plot(self.retention_time[~MS1], NP[~MS1], label="MS2")
        plt.legend()
        plt.xlabel("Retention Time")
        plt.ylabel("Number of Peaks")
        plt.title("Peak Counts per Frame")
        if show:
            plt.show()


    def intensity_per_frame(self, recalibrated=True):
        """Get sum of intensity per each frame (retention time).

        Arguments:
            recalibrated (bool): Use Bruker recalibrated total intensities or calculate them from scratch with OpenTIMS?
        
        Returns:
            np.array: sums of intensities per frame. 
        """
        return self.frames.SummedIntensities.values if recalibrated else self.framesTIC()


    def plot_TIC(self, recalibrated=True, show=True):
        """Plot peak counts per frame.

        Arguments:
            recalibrated (bool): Use Bruker recalibrated total intensities or calculate them from scratch with OpenTIMS?
            show (bool): Show the plot immediately, or just add it to the canvas?
        """
        import matplotlib.pyplot as plt
        MS1 = self._ms1_mask
        I = self.intensity_per_frame(recalibrated)
        plt.plot(self.retention_time[ MS1], I[ MS1], label="MS1")
        plt.plot(self.retention_time[~MS1], I[~MS1], label="MS2")
        plt.legend()
        plt.xlabel("Retention Time")
        plt.ylabel("Intensity")
        plt.title("Total Intensity [Ion Current]")
        if show:
            plt.show()


    #TODO: this should be reimplemented later on in C++, single core...
    def intensity_given_mz_inv_ion_mobility(self,
                                            frames=None,
                                            mz_bin_borders=np.linspace(500, 2500, 1001),
                                            inv_ion_mobility_bin_borders=np.linspace(0.8, 1.7, 101)):
        """Sum intensity over m/z-inverse ion mobility rectangles.

        Typically it does not make too much sense to mix MS1 intensities with the others here.

        Arguments:
            frames (iterable): Frames to consider. Defaults to all ms1_frames. 
            mz_bin_borders (np.array): Positions of bin borders for mass over charge ratios.
            inv_ion_mobility_bin_borders (np.array): Positions of bin borders for inverse ion mobilities.
        Returns:
            tuple: np.array with intensities, the positions of bin borders for mass over charge ratios and inverse ion mobilities.
        """
        if frames is None:
            frames = self.ms1_frames

        I = np.zeros(shape=(len(mz_bin_borders)-1,
                            len(inv_ion_mobility_bin_borders)-1),
                     dtype=float)
        # float because numpy does not have histogram2d with ints 

        for X in self.query_iter(frames=frames,
                                 columns=('mz','inv_ion_mobility','intensity')):
            I_fr, _,_ = np.histogram2d(X.mz, X.inv_ion_mobility,
                                       bins=[mz_bin_borders,
                                             inv_ion_mobility_bin_borders], 
                                       weights=X.intensity)
            I += I_fr

        return I, mz_bin_borders, inv_ion_mobility_bin_borders


    def plot_intensity_given_mz_inv_ion_mobility(self, 
                                   intensity_transformation=np.sqrt,
                                   show=True,
                                   imshow_kwds={'interpolation':'nearest',
                                                'aspect':'auto'},
                                   **kwds):
        """Sum intensity over m/z-inverse ion mobility rectangles.

        Plot a transformation of the sum of intensities.
        Usually, plotting the square root of summed intensities looks best.


        Arguments:
            intensity_transformation (np.ufunc): Function that transforms intensities. Default to square root.
            show (bool): Show the plot immediately, or just add it to the canvas?
            **kwds: Keyword arguments for the 'intensity_given_mz_inv_ion_mobility' method.
        """
        import matplotlib.pyplot as plt
        
        I, mz_bin_borders, inv_ion_mobility_bin_borders = self.intensity_given_mz_inv_ion_mobility(**kwds)
        plt.imshow(intensity_transformation(I),
                   extent=[mz_bin_borders[0], mz_bin_borders[-1],
                           inv_ion_mobility_bin_borders[0], inv_ion_mobility_bin_borders[-1]],
                   **imshow_kwds)
        plt.xlabel("Mass / Charge")
        plt.ylabel("Inverse Ion Mobility")
        plt.title("Total Intensity")
        if show:
            plt.show()


    def to_hdf(self,
               target_path,
               columns=all_columns,
               compression='gzip',
               compression_level=9,
               shuffle=True,
               chunks=True,
               silent=True,
               **kwds):
        """Convert the data set to HDF5 compatible with 'vaex'.

        Most of the arguments are documented on the h5py website.

        Arguments:
            target_path (str): Where to write the file (folder will be automatically created). Cannot point to an already existing file.
            columns (tuple): Names of columns to export to HDF5.
            compression (str): Compression strategy.
            compression_level (str): Parameters for compression filter.
            shuffle (bool): Enable shuffle filter.
            chunks (int): Chunk shape, or True to enable auto-chunking.
            silent (bool): Skip progress bar
        """
        import h5py

        target_path = pathlib.Path(target_path)
        target_path.parent.mkdir(parents=True, exist_ok=True)

        with h5py.File(target_path, "w") as hdf_conn:
            out_grp = hdf_conn.create_group("data")

            datasets = {}
            for colname in columns:
                datasets[colname] = out_grp.create_dataset(
                    name=colname,
                    shape=(len(self),),
                    compression=compression,
                    compression_opts=compression_level if compression=='gzip' else None,
                    dtype=column2dtype[colname],
                    chunks=chunks,
                    shuffle=shuffle,
                    **kwds)

            frame_ids = range(self.min_frame, self.max_frame+1)
            if not silent:
                frame_ids = tqdm.tqdm(frame_ids)

            data_offset = 0
            for frame_id in frame_ids:
                # super for compatibility with Michal's code...
                frame = super().query(frame_id, columns=columns)
                frame_size = len(next(frame.values().__iter__()))
                for colname, dataset in datasets.items():
                    dataset.write_direct(frame[colname], dest_sel=np.s_[data_offset:data_offset+frame_size])
                data_offset += frame_size

        if not silent:
            print(f"Finished with {target_path}")
